import logging
import traceback
import pandas as pd
import numpy as np
import seaborn as sns
from collections import defaultdict

import matplotlib
matplotlib.use('Agg')
matplotlib.rcParams['pdf.fonttype'] = 42

import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.plotting.utilities
from inStrain.plotting.utilities import plot_genome
from inStrain.plotting.utilities import estimate_breadth

def genome_plot_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        stb = IS.get('scaffold2bin')
        b2s = defaultdict(list)
        for s, b in stb.items():
            b2s[b].append(s)
        assert len(b2s.keys()) > 0

        # Load the cache
        covTs = kwargs.get('covT')#, IS.get('covT'))
        clonTs = kwargs.get('clonT')#, IS.get('clonT'))
        raw_linkage_table = kwargs.get('raw_linkage_table')#, IS.get('raw_linkage_table'))
        cumulative_snv_table = kwargs.get('cumulative_snv_table')#, IS.get('cumulative_snv_table'))
        scaffold2length = IS.get('scaffold2length')
        rl = IS.get_read_length()
        profiled_scaffolds = set(scaffold2length.keys())

    except:
        logging.error("Skipping plot 2 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 2")
    name = 'genomeWide_microdiveristy_metrics.pdf'
    pp = PdfPages(plot_dir + name)


    for genome, scaffolds in b2s.items():
        if not plot_genome(genome, IS, **kwargs):
            continue
        present_scaffolds = list(set(scaffolds).intersection(set(profiled_scaffolds)))
        Wdb, breaks, midpoints = load_windowed_metrics(present_scaffolds,
                                scaffold2length,
                                rl,
                                report_midpoints=True,
                                covTs=covTs, clonTs=clonTs,
                                raw_linkage_table=raw_linkage_table,
                                cumulative_snv_table=cumulative_snv_table)
        if len(Wdb) == 0:
            logging.debug(f"{genome} could not have windowed metrics loaded")
            continue
        genomeWide_microdiveristy_metrics_plot(Wdb, breaks, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(8, 5)
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def scaffold_inspection_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        stb = IS.get('scaffold2bin')
        b2s = defaultdict(list)
        for s, b in stb.items():
            b2s[b].append(s)
        assert len(b2s.keys()) > 0

        # Load the cache
        covTs = kwargs.get('covTs', IS.get('covT'))
        clonTs = kwargs.get('clonTs', IS.get('clonT'))
        raw_linkage_table = kwargs.get('raw_linkage_table', IS.get('raw_linkage_table'))
        cumulative_snv_table = kwargs.get('cumulative_snv_table', IS.get('cumulative_snv_table'))
        scaffold2length = IS.get('scaffold2length')
        rl = IS.get_read_length()
        profiled_scaffolds = set(scaffold2length.keys())

    except:
        logging.error("Skipping plot 7 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 7")
    name = 'ScaffoldInspection_plot.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, scaffolds in b2s.items():
        if not plot_genome(genome, IS, **kwargs):
            continue
        present_scaffolds = list(set(scaffolds).intersection(set(profiled_scaffolds)))
        Wdb, breaks, midpoints = load_windowed_metrics(present_scaffolds,
                                scaffold2length,
                                rl,
                                report_midpoints=True,
                                covTs=covTs, clonTs=clonTs,
                                raw_linkage_table=raw_linkage_table,
                                cumulative_snv_table=cumulative_snv_table)
        if len(Wdb) == 0:
            logging.debug(f"{genome} could not have windowed metrics loaded")
            continue
        scaffold_inspection_plot(Wdb, breaks, midpoints, title=genome)
        fig = plt.gcf()
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def genomeWide_microdiveristy_metrics_plot(Wdb, breaks, title=''):
    '''
    Make the multiple metrics plot
    '''
    # Get set up for multiple rows
    i = len(Wdb['metric'].unique())
    if i > 1:
        fig, ax = plt.subplots(i, 1, sharex=True)
    else:
        ax = {}
        ax[0] = plt.gca()

    i = 0
    for metric in ['linkage', 'snp_density', 'coverage', 'nucl_diversity']:
    #for metric, wdb in Wdb.groupby('metric'):
        if metric not in set(Wdb['metric'].tolist()):
            continue

        wdb = Wdb[Wdb['metric'] == metric]
        med = wdb['value'].median()

        # Rotate colors:
        colors = ['red', 'blue', 'black']
        c = 0
        for mm, ddb in wdb.groupby('ANI'):
            ax[i].plot(ddb['midpoint'], ddb['value'], c=colors[c], label=mm, marker='o', ms=1)#, ls='')
            c += 1

        ax[i].set_title("{0}".format(metric))
        ax[i].grid(False)

        if i == 0:
            ax[i].legend(loc='upper left', title='Min read ANI (%)')

        # Add breaks
        for b in breaks:
            ax[i].axvline(b, ls='-', c='lightgrey', zorder=-1)

        i += 1

    plt.xlabel('genome position')
    plt.xlim(0, Wdb['midpoint'].max())
    plt.suptitle(title, y=0.999)
    plt.subplots_adjust(hspace=0.3)

def load_windowed_metrics(scaffolds, s2l, rLen, metrics=None, window_len=None, ANI_levels=[0, 100],
                        min_scaff_len=0, report_midpoints=False, covTs=False, clonTs=False,
                        raw_linkage_table=False, cumulative_snv_table=False):

    if metrics is None:
        metrics = ['coverage', 'nucl_diversity', 'linkage', 'snp_density']

    if type(metrics) != type([]):
        print("Metrics must be a list")
        return

    # Figure out the MMs needed
    #rLen = IS.get_read_length()
    mms = [_get_mm(None, ANI, rLen=rLen) for ANI in ANI_levels]

    # Sort the scaffolds
    #s2l = IS.get('scaffold2length')
    scaffolds = sorted(scaffolds, key=s2l.get, reverse=True)

    if min_scaff_len > 0:
        scaffolds = [s for s in scaffolds if s2l[s] >= min_scaff_len]

    # Figure out the window length
    if window_len == None:
        window_len = int(sum([s2l[s] for s in scaffolds]) / 100)
    else:
        window_len = int(window_len)

    # Calculate the breaks
    breaks = []
    midpoints = {}
    tally = 0
    for scaffold in scaffolds:
        midpoints[scaffold] = tally + int(s2l[scaffold] / 2)
        tally += s2l[scaffold]
        breaks.append(tally)


    dbs = []
    if 'coverage' in metrics:
        if covTs == False:
            logging.error("need covTs for coverage")
            raise Exception
        cdb = load_windowed_coverage_or_clonality('coverage', covTs, scaffolds, window_len, mms, ANI_levels, s2l)
        cdb['metric'] = 'coverage'
        dbs.append(cdb)
#     if 'clonality' in metrics:
#         cdb = load_windowed_coverage_or_clonality(IS, 'clonality', scaffolds, window_len, mms, ANI_levels, s2l)
#         cdb['metric'] = 'clonality'
#         dbs.append(cdb)
    if 'nucl_diversity' in metrics:
        if clonTs == False:
            logging.error("need clonTs for microdiversity")
            raise Exception
        cdb = load_windowed_coverage_or_clonality('nucl_diversity', clonTs, scaffolds, window_len, mms, ANI_levels, s2l)
        cdb['metric'] = 'nucl_diversity'
        dbs.append(cdb)
    if 'linkage' in metrics:
        if raw_linkage_table is False:
            logging.error("need raw_linkage_table for linkage")
            raise Exception
        cdb = load_windowed_linkage(raw_linkage_table, scaffolds, window_len, mms, ANI_levels, s2l)
        cdb['metric'] = 'linkage'
        dbs.append(cdb)
    if 'snp_density' in metrics:
        if cumulative_snv_table is False:
            logging.error("need cumulative_snv_table for snp_density")
            raise Exception
        if len(cumulative_snv_table) > 0:
            cdb = load_windowed_SNP_density(cumulative_snv_table, scaffolds, window_len, mms, ANI_levels, s2l)
            cdb['metric'] = 'snp_density'
            dbs.append(cdb)

    if len(dbs) > 0:
        Wdb = pd.concat(dbs, sort=True)
        Wdb = Wdb.rename(columns={'avg_cov':'value'})
    else:
        Wdb = pd.DataFrame()

    # Add blanks at the breaks
    table = defaultdict(list)
    for mm, ani in zip(mms, ANI_levels):
        for metric in Wdb['metric'].unique():
            for bre in breaks:
                table['scaffold'].append('break')
                table['mm'].append(mm)
                table['ANI'].append(ani)
                table['adjusted_start'].append(bre) # The minus one makes sure it doenst split things it shouldnt
                table['adjusted_end'].append(bre)
                table['value'].append(np.nan)
                table['metric'].append(metric)
    bdb = pd.DataFrame(table)
    Wdb = pd.concat([Wdb, bdb], sort=False)
    if len(Wdb) > 0:
        Wdb['midpoint'] = [np.mean([x, y]) for x, y in zip(Wdb['adjusted_start'], Wdb['adjusted_end'])]
        Wdb = Wdb.sort_values(['metric', 'mm', 'midpoint', 'scaffold'])

    if report_midpoints:
        return Wdb, breaks, midpoints
    else:
        return Wdb, breaks

def _get_mm(IS, ANI, rLen = None):
    '''
    Get the mm corresponding to an ANI level in an IS
    '''
    if ANI > 1:
        ANI = ANI / 100

    if rLen == None:
        rLen = IS.get_read_length()
        #rLen = IS.get('mapping_info')['mean_pair_length'].tolist()[0]
    mm = int(round((rLen - (rLen * ANI))))
    return mm

def load_windowed_coverage_or_clonality(thing, covTs, scaffolds, window_len, mms, ANI_levels, s2l):
    '''
    Get the windowed coverage

    Pass in a clonTs for microdiversity and covTs for coverage
    '''
    if thing == 'coverage':
        item = 'covT'
    elif thing == 'nucl_diversity':
        item = 'clonT'
    else:
        print("idk what {0} is".format(thing))
        return

    # Get the covTs
    #covTs = IS.get(item, scaffolds=scaffolds)

    # Make the windows
    dbs = []
    tally = 0
    breaks = []
    for scaffold in scaffolds:
        if scaffold not in covTs:
            tally += s2l[scaffold]
            breaks.append(tally)
            continue
        else:
            covT = covTs[scaffold]

        for mm, ani in zip(mms, ANI_levels):
            if item == 'covT':
                cov = inStrain.profile.profile_utilities.mm_counts_to_counts_shrunk(covT, mm)
                if len(cov) == 0:
                    continue
                db = _gen_windowed_cov(cov, window_len, sLen=s2l[scaffold])

            elif item == 'clonT':
                cov = _get_basewise_clons3(covT, mm)
                if len(cov) == 0:
                    continue
                db = _gen_windowed_cov(cov, window_len, sLen=s2l[scaffold], full_len=False)
                db['avg_cov'] = [1 - x if x == x else x for x in db['avg_cov']]

            db['scaffold'] = scaffold
            db['mm'] = mm
            db['ANI'] = ani

            db['adjusted_start'] = db['start'] + tally
            db['adjusted_end'] = db['end'] + tally

            dbs.append(db)
        tally += s2l[scaffold]
        breaks.append(tally)

    if len(dbs) > 0:
        Wdb = pd.concat(dbs)
    else:
        Wdb = pd.DataFrame()
    return Wdb#, breaks

def load_windowed_linkage(Ldb, scaffolds, window_len, mms, ANI_levels, s2l, on='r2'):

    # Get the linkage table
    #Ldb = IS.get('raw_linkage_table')
    Ldb = Ldb[Ldb['scaffold'].isin(scaffolds)].sort_values('mm')
    got_scaffolds = set(Ldb['scaffold'].unique())

    # Make the windows
    dbs = []
    tally = 0
    breaks = []
    for scaffold in scaffolds:
        if scaffold not in got_scaffolds:
            tally += s2l[scaffold]
            breaks.append(tally)
            continue
        else:
            ldb = Ldb[Ldb['scaffold'] == scaffold]

        for mm, ani in zip(mms, ANI_levels):
            db = ldb[ldb['mm'] <= int(mm)].drop_duplicates(subset=['scaffold', 'position_A', 'position_B'], keep='last')
            cov = db.set_index('position_A')[on].sort_index()

            db = _gen_windowed_cov(cov, window_len, sLen=s2l[scaffold], full_len=False)
            db['scaffold'] = scaffold
            db['mm'] = mm
            db['ANI'] = ani

            db['adjusted_start'] = db['start'] + tally
            db['adjusted_end'] = db['end'] + tally

            dbs.append(db)
        tally += s2l[scaffold]
        breaks.append(tally)

    if len(dbs) > 0:
        Wdb = pd.concat(dbs)
    else:
        Wdb = pd.DataFrame()
    return Wdb

def load_windowed_SNP_density(Ldb, scaffolds, window_len, mms, ANI_levels, s2l):
    # Get the table
    #Ldb = IS.get('cumulative_snv_table')
    Ldb = Ldb[Ldb['scaffold'].isin(scaffolds)].sort_values('mm')

    got_scaffolds = list(Ldb['scaffold'].unique())

    # Make the windows
    dbs = []
    tally = 0
    breaks = []
    for scaffold in scaffolds:
        if scaffold not in got_scaffolds:
            tally += s2l[scaffold]
            breaks.append(tally)
            continue
        else:
            ldb = Ldb[Ldb['scaffold'] == scaffold]

        for mm, ani in zip(mms, ANI_levels):
            db = ldb[ldb['mm'] <= int(mm)].drop_duplicates(subset=['scaffold', 'position'], keep='last')
            cov = db.set_index('position')['ref_base'].sort_index()

            db = _gen_windowed_cov(cov, window_len, sLen=s2l[scaffold], full_len='count')
            db['scaffold'] = scaffold
            db['mm'] = mm
            db['ANI'] = ani

            db['adjusted_start'] = db['start'] + tally
            db['adjusted_end'] = db['end'] + tally

            dbs.append(db)
        tally += s2l[scaffold]
        breaks.append(tally)

    if len(dbs) > 0:
        Wdb = pd.concat(dbs)
    else:
        Wdb = pd.DataFrame()
    return Wdb

def _gen_windowed_cov(cov, window_len, sLen=False, full_len=True):
    '''
    From a series of coverage values, return windows
    '''
    if sLen == False:
        sLen = cov.index.max()

    table = defaultdict(list)
    i = 0
    numberChunks = sLen // window_len + 1
    for cov in iterate_chunks_series(cov, chunkSize=window_len, sLen=sLen):
        if i + 1 == numberChunks:
            mLen = sLen - (i * window_len)
            if mLen == 0:
                continue
        else:
            mLen = window_len

        table['start'].append(i*window_len)
        table['end'].append(i*window_len + mLen)

        if full_len == True:
            table['avg_cov'].append(cov.sum() / mLen)
        elif full_len == False:
            table['avg_cov'].append(cov.mean())
        elif full_len == 'count':
            table['avg_cov'].append(len(cov)/mLen)
        i += 1

    return pd.DataFrame(table)

def _get_basewise_clons3(clonT, MM, fill_zeros=False):
    p2c = {}
    mms = sorted([int(mm) for mm in list(clonT.keys()) if int(mm) <= int(MM)])
    for mm in mms:
        p2c.update(clonT[mm].to_dict())

    inds = []
    vals = []
    for ind in sorted(p2c.keys()):
        inds.append(ind)
        vals.append(p2c[ind])

    counts = pd.Series(data = vals, index = np.array(inds).astype('int'))

    if fill_zeros:
        counts = counts.append(pd.Series(np.zeros(fill_zeros - len(counts))))

    return counts

def iterate_chunks_series(d, chunkSize=100, sLen=False):
    '''
    Break up Ndbs into chunks
    '''
    if sLen == False:
        sLen = d.index.max()

    numberChunks = sLen // chunkSize + 1
    for i in range(numberChunks):
        #print(i, int(i*chunkSize), int((i+1)*chunkSize), len(d))
        start = int(i*chunkSize)
        end = int((i+1)*chunkSize)
        yield (d.loc[start:end])

def scaffold_inspection_plot(Wdb, breaks, midpoints, title=''):
    '''
    Make the multiple metrics plot
    '''
    sns.set_style('whitegrid')

    # Get set up for multiple rows
    i = len(Wdb['metric'].unique())
    if i > 1:
        fig, ax = plt.subplots(1, i, sharey=True)
    else:
        ax = {}
        ax[0] = plt.gca()

    i = 0
    for metric in ['linkage', 'snp_density', 'coverage', 'nucl_diversity']:
    #for metric, wdb in Wdb.groupby('metric'):
        if metric not in set(Wdb['metric'].tolist()):
            continue

        wdb = Wdb[Wdb['metric'] == metric]

        # Rotate colors:
        colors = ['red', 'blue', 'black']
        c = 0
        for mm, ddb in wdb.groupby('ANI'):
            ax[i].plot(ddb['value'], ddb['midpoint'], c=colors[c], label=mm, marker='o', ms=5)#, ls='')
            c += 1

        ax[i].set_title("{0}".format(metric))
        ax[i].yaxis.grid(False)
        #ax[i].grid(False)
        #ax[i].set_ylim(Wdb['midpoint'].max(), 0)
        #ax[i].set_xlim([0, med*2])

        if i == 0:
            ax[i].legend(loc='upper left', title='Min read ANI (%)')

        # Add breaks
        for b in breaks:
            ax[i].axhline(b, ls='-', c='lightgrey', zorder=-1)

        i += 1

    # Annotate the scaffolds
    locations = []
    scaffolds = []
    for scaff, l in midpoints.items():
        locations.append(l)
        scaffolds.append(scaff)
    plt.yticks(locations, scaffolds, fontsize=1)
    plt.ylim(Wdb['midpoint'].max(), 0)

    plt.suptitle(title, y=0.999)
    plt.subplots_adjust(hspace=0.3)
    #plt.gca().xaxis.grid(False)

    fig = plt.gcf()
    ylim = inStrain.plotting.utilities._calc_ylim(Wdb['midpoint'].max())
    fig.set_size_inches(8, ylim)


"""
This handles the argument parsing and logic for profiling bam files
"""

# Import packages
import os
import copy
import logging
import pandas as pd
from tqdm import tqdm
import multiprocessing
import traceback
from collections import defaultdict
from tqdm import tqdm

import inStrain.readComparer
import inStrain.compare_utils
import inStrain.genomeUtilities
import inStrain.controller

class CompareController(object):
    '''
    Main controller of the profile command
    '''
    def __init__(self, args):
        '''
        Set all of the command line arguments in the "args" attribute

        Doing it this way lets your pass the arguments to other controllers
        '''
        self.args = args
        self.ori_args = copy.deepcopy(args)
        self.kwargs = vars(self.args)

        # Make null model for SNP calling
        fdr = self.kwargs.get('fdr', 1e-6)
        null_loc = os.path.dirname(__file__) + '/helper_files/NullModel.txt'
        self.null_model = inStrain.profile.snv_utilities.generate_snp_model(null_loc, fdr=fdr)

    def main(self):
        '''
        The main method when run on the command line
        '''
        self.parse_arguments()

        # Set up which scaffolds will be compared when
        message = """\
***************************************************
    ..:: inStrain compare Step 1. Load data ::..
***************************************************
                """
        logging.info(message)
        self.create_scaffoldcomparison_objects()

        # Actually do the processing
        message = """\
***************************************************
..:: inStrain compare Step 2. Run comparisons ::..
***************************************************
                """
        logging.info(message)
        self.run_comparisons()

        # Do auxillary processing if needed
        message = """\
***************************************************
..:: inStrain compare Step 3. Auxiliary processing ::..
***************************************************
                """
        logging.info(message)
        self.run_auxillary_processing()

        # Store the results
        message = """\
***************************************************
..:: inStrain compare Step 4. Store results ::..
***************************************************
                """
        logging.info(message)
        self.store_results()
        self.write_final_message()

    def parse_arguments(self):
        """
        Parse the arguments and add them this object as attributes
        """
        args = self.args

        # Set up output object and log
        outbase = args.output
        RCprof = inStrain.SNVprofile.SNVprofile(outbase)
        log_loc = RCprof.get_location('log') + 'log.log'
        inStrain.controller.setup_logger(log_loc)
        self.RCprof = RCprof

        # Set up list of input IS profiles
        inputs = list(args.input)
        assert len(inputs) > 1, "You need to have more than one input .IS file"
        self.inputs = inputs

        # Set up a list of scaffolds to compare
        if args.scaffolds is not None:
            self.scaffolds_to_compare = inStrain.profile.fasta.load_scaff_list(args.scaffolds)
        else:
            self.scaffolds_to_compare = None

        # Load the scaffold to bin file
        if args.stb is not None:
            if len(args.stb) > 0:
                # Load it
                stb = inStrain.genomeUtilities.load_scaff2bin(args.stb)
                self.stb = stb

                # Only compare scaffolds in the .stb
                if self.scaffolds_to_compare is not None:
                    self.scaffolds_to_compare = self.scaffolds_to_compare.union(set(stb.keys()))
                else:
                    self.scaffolds_to_compare = set(stb.keys())

        # Load database mode results
        if self.args.database_mode is True:
            inStrain.logUtils.log_checkpoint("Compare", "LoadDatabaseMode", "start")

            bin2scaffolds = defaultdict(list)
            for s, b in stb.items():
                bin2scaffolds[b].append(s)

            input2scaffolds = {}
            for input in self.inputs:
                scaffolds = inStrain.compare_utils.find_relevant_scaffolds(input, bin2scaffolds, self.kwargs)
                input2scaffolds[input] = scaffolds

            self.input2scaffolds = input2scaffolds
            inStrain.logUtils.log_checkpoint("Compare", "LoadDatabaseMode", "end")

        # Handle single genome mode
        if self.args.genome is not None:
            bin2scaffolds = defaultdict(list)
            for s, b in stb.items():
                bin2scaffolds[b].append(s)
            if self.args.genome in bin2scaffolds:
                self.scaffolds_to_compare = self.scaffolds_to_compare.intersection(set(bin2scaffolds[self.args.genome]))
            else:
                logging.error(f'genome {self.args.genome} is not in the provided .stb file')
                raise Exception

    def create_scaffoldcomparison_objects(self):
        """
        Figure out which scaffolds need to be compared
        """
        # Load scaffolds to compare
        if self.scaffolds_to_compare is not None:
            scaffolds_to_compare = set(self.scaffolds_to_compare)
        else:
            scaffolds_to_compare = None

        # Load ScaffoldComparison objects
        if self.args.database_mode is True:
            valid_SCs, scaffold2length = make_scaffoldcomparison_objects(self.inputs, scaffolds_to_compare,
                                                                         input2scaffolds=self.input2scaffolds)
        else:
            valid_SCs, scaffold2length = make_scaffoldcomparison_objects(self.inputs, scaffolds_to_compare)

        self.SC_objects = valid_SCs
        self.scaffold2length = scaffold2length

        # Establish ScaffoldComparison groups
        group_length = self.kwargs.get('group_length', 10000000)

        #Cdb = calc_scaff_sim_matrix(valid_SCs)
        #SC_groups = establish_SC_groups(valid_SCs, Cdb, group_length)
        SC_groups = group_Scaffold_objects(valid_SCs, group_length)

        self.scaffold_comparison_groups = SC_groups

    def run_comparisons(self):
        """
        Run each group one at a time, storing the results in the meantime
        """
        cdbs = [] # Store scaffold comparison information
        mdbs = [] # Store mismatch locations
        pair2mm2covOverlaps = [] # Store coverage overlap locations
        order = [] # Store order of scaffolds

        groups = len(self.scaffold_comparison_groups)
        for i, SCgroup in enumerate(self.scaffold_comparison_groups):
            logging.info(f'Running group {i+1} of {groups}')
            SCgroup.load_cache()
            results = inStrain.compare_utils.run_compare_multiprocessing(SCgroup.cmd_queue, SCgroup.result_queue,
                                                                         self.null_model, num_to_run=len(SCgroup.scaffolds),
                                                                         **self.kwargs)
            for result in results:
                if result is not None:
                    Cdb, Mdb, pair2mm2covOverlap, scaffold = result
                    for item, lis in zip([Cdb, Mdb, pair2mm2covOverlap, scaffold], [cdbs, mdbs, pair2mm2covOverlaps, order]):
                        lis.append(item)

            SCgroup.purge_cache()

        # Process results
        self.process_results(cdbs, mdbs, pair2mm2covOverlaps, order)

    def process_results(self, cdbs, mdbs, pair2mm2covOverlaps, order):
        """
        Merge and store results
        """
        if len(pair2mm2covOverlaps) > 0:
            scaff2pair2mm2overlap = {}
            for scaff, pair2mm2covOverlap in zip(order, pair2mm2covOverlaps):
                scaff2pair2mm2overlap[scaff] = pair2mm2covOverlap
        else:
            scaff2pair2mm2overlap = None

        if len(mdbs) > 0:
            Mdb = pd.concat(mdbs, sort=False)
        else:
            Mdb = None

        # Do some typing of this DataFrame
        if len(Mdb) > 0:
            # These should never be NA
            int_cols = ['position', 'mm']
            for c in int_cols:
                Mdb[c] = Mdb[c].astype(int)

            # Handle bools
            bool_cols = ['consensus_SNP', 'population_SNP']
            for c in bool_cols:
                Mdb[c] = Mdb[c].astype(bool)

        self.comparison_db = pd.concat(cdbs, sort=False)
        self.mismatch_location_db = Mdb
        self.scaff2pair2mm2overlap = scaff2pair2mm2overlap

    def run_auxillary_processing(self):
        """
        Handle .stb files, database mode, and more
        """
        # Calculate s2l (scaffold 2 length)
        s2l = {}
        for SC in self.SC_objects:
            s2l[SC.scaffold] = SC.length
        self.s2l = s2l

        # Calculate genome-level results
        if hasattr(self, 'stb'):
            # Calculate bin 2 length
            # Make bin to length
            b2l = {}
            for scaffold, bin in self.stb.items():
                if bin not in b2l:
                    b2l[bin] = 0

                if scaffold in s2l:
                    b2l[bin] += s2l[scaffold]
                else:
                    logging.debug(
                        "FAILURE StbError {0} {1} no_length will not be considered as part of the genome".format(
                            scaffold, bin))
            self.bin2length = b2l

            gdb = inStrain.genomeUtilities._add_stb(self.comparison_db, self.stb)
            Gdb = inStrain.genomeUtilities._genome_wide_readComparer(gdb, self.stb, b2l, **self.kwargs)
            self.genomelevel_compare = Gdb

        # Cluster genomes
        if hasattr(self, 'genomelevel_compare'):
            self.run_genome_clustering()

    def store_results(self):
        # Store the results in the RC
        inStrain.logUtils.log_checkpoint("Compare", "SaveResults", "start")

        self.RCprof.store('comparisonsTable', self.comparison_db, 'pandas', 'Comparisons between the requested IS objects')
        self.RCprof.store('scaffold2length', self.s2l, 'dictionary', 'Scaffold to length')

        # Store auxillary things
        if hasattr(self, 'bin2length'):
            self.RCprof.store('bin2length', self.bin2length, 'dictionary', 'Dictionary of bin 2 total length')
        if hasattr(self, 'genomelevel_compare'):
            # ... this is jankey, but I guess it's what we do
            out_base = self.RCprof.get_location('output') + os.path.basename(self.RCprof.get('location')) + '_'
            self.genomelevel_compare.to_csv(out_base + 'genomeWide_compare.tsv', index=False, sep='\t')
        if hasattr(self, 'Cdb'):
            # ... this is jankey, but I guess it's what we do
            out_base = self.RCprof.get_location('output') + os.path.basename(self.RCprof.get('location')) + '_'
            self.Cdb.to_csv(out_base + 'strain_clusters.tsv', index=False, sep='\t')
        if hasattr(self, 'stb'):
            self.RCprof.store('scaffold2bin', self.stb, 'dictionary', 'Dictionary of scaffold 2 bin')

        # Make the output files
        self.RCprof.generate('comparisonsTable')

        # Store scaff2pair2mm2SNPs
        if self.args.store_mismatch_locations:
            self.RCprof.store('pairwise_SNP_locations', self.mismatch_location_db, 'pandas',
                         'A dataframe of scaffold, IS pair, mm level, SNP locations')
            self.RCprof.generate('pairwise_SNP_locations')

        # Store scaff2pair2mm2cov
        if self.args.store_coverage_overlap:
            self.RCprof.store('scaff2pair2mm2cov', self.scaff2pair2mm2overlap, 'special',
                         'A dictionary of scaffold -> IS pair -> mm level -> positions with coverage overlap')

        inStrain.logUtils.log_checkpoint("Compare", "SaveResults", "end")

        # Make plots
        if hasattr(self, 'genomelevel_compare'):
            self.make_plots()

    def make_plots(self):
        '''
        Call plotting function from "profile" module
        '''
        args = self.args
        args.IS = self.RCprof.location

        if not args.skip_plot_generation:
            inStrain.logUtils.log_checkpoint("Compare", "making_plots", "start")
            args.plots = ['10']
            inStrain.controller.Controller().plot_operation(args)
            inStrain.logUtils.log_checkpoint("Compare", "making_plots", "end")
        else:
            logging.info('Nevermind! You chose to skip making plots')

    def write_final_message(self):
        Sprofile = self.RCprof
        message = """\
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

..:: inStrain compare finished ::..

Output tables........ {0}
Figures.............. {1}
Logging.............. {2}

See documentation for output descriptions - https://instrain.readthedocs.io/en/latest/

$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
        """.format(Sprofile.get_location('output'), \
                   Sprofile.get_location('figures'),
                   Sprofile.get_location('log'))
        logging.info(message)

    def run_genome_clustering(self):
        """
        Cluster genomes dRep style
        """
        kwargs = self.kwargs
        Mdb = self.genomelevel_compare
        Mdb = Mdb.sort_values(['genome', 'name1', 'name2'])
        if len(Mdb) == 0:
            return

        try:
            Cdb = inStrain.compare_utils.cluster_genome_strains(Mdb, kwargs)
            self.Cdb = Cdb
        except:
            logging.error('Could not cluster genomes; heres a traceback:')
            traceback.print_exc()

class ScaffoldComparison(object):
    """
    Holds the information needed to run a scaffold comparison and the results of said comparison
    """
    def __init__(self, scaffold, length):
        """
        scaffold = name of the scaffold
        length = length of scaffold
        profiles = list of ISP profiles with this scaffold (same order as names)
        names = list of names of ISP profiles with this scaffold (same order as profiles)
        """
        self.scaffold = scaffold
        self.length = int(length)

        self.profiles = []
        self.names = []

        self.SNPtables = []
        self.covTs = []

    def add_profile(self, ISP, name):
        """
        Add an ISP object with this scaffold that should be compared
        """
        self.profiles.append(ISP)
        self.names.append(name)

    def valid(self):
        """
        Return True if this scaffold has more than 2 profiles in it that can be compared
        """
        assert len(self.names) == len(set(self.names)), "You're comparing ISP profiles that have the same names. No no."
        return len(self.profiles) > 1

    def compare(self, OSC):
        """
        Return the similarity of this scaffold to the other, based on overname of names
        """
        n1 = set(self.names)
        n2 = set(OSC.names)
        return len(n1.intersection(n2)) / len(n1.union(n2))

class ScaffoldCompareGroup(object):
    """
    Holds a group of scaffolds that will be parallelized together. Ideally these scaffolds should have a high overlap
    in the profiles that they are contained in
    """
    def __init__(self, SCs):
        """
        SCs = list of scaffold comparison objects
        """
        self.ScaffoldComparisons = SCs

        # Establish the ISPs and names associated with this group of scaffolds
        ISPs = []
        names = []
        scaffolds = []
        for sc in SCs:
            scaffolds.append(sc.scaffold)
            for isp, name in zip(sc.profiles, sc.names):
                if name not in names:
                    ISPs.append(isp)
                    names.append(name)

        self.ISPs = ISPs
        self.names = names
        self.scaffolds = scaffolds

    def load_cache(self):
        """
        Load information from ISPs from disk to this object
        """
        names = self.names
        sProfiles = self.ISPs
        scaffolds_to_compare = set(self.scaffolds)

        # Make sure no duplicates in names
        assert len(names) == len(set(names)), 'Cannot have 2 of the same named IS {0}'.format(names)

        # Load data from the profiles
        name2covT = {}
        name2SNPtable = {}

        for S, name in zip(sProfiles, names):
            name2covT[name] = S.get('covT', scaffolds=scaffolds_to_compare)
            name2SNPtable[name] = S.get('cumulative_snv_table').rename(
                columns={'conBase': 'con_base', 'refBase': 'ref_base', 'varBase': 'var_base',
                         'baseCoverage': 'position_coverage'})

        # Attach this information to ScaffoldComparison objects
        for SC in self.ScaffoldComparisons:
            for name in SC.names:
                SC.SNPtables.append(inStrain.compare_utils.subset_SNP_table(name2SNPtable[name], SC.scaffold))
                SC.covTs.append(name2covT[name][SC.scaffold])

        # Set up a command and result queue
        ctx = multiprocessing.get_context('spawn')
        self.cmd_queue = ctx.Queue()
        self.result_queue = ctx.Queue()
        for SC in self.ScaffoldComparisons:
            self.cmd_queue.put(SC)

    def purge_cache(self):
        """
        Purge information from ISPs from RAM
        """
        for SC in self.ScaffoldComparisons:
            del SC.SNPtables
            del SC.covTs

    def __str__(self):
        """
        String representation
        """
        string = f"This group has {len(self.ScaffoldComparisons)} scaffolds accross {len(self.names)} ISP objects"
        return string

def group_Scaffold_objects(valid_SCs, group_length):
    SC_groups = []
    current_group = []
    current_size = 0
    for s in valid_SCs:
        current_group.append(s)
        current_size += s.length
        if current_size >= group_length:
            SC_groups.append(ScaffoldCompareGroup(current_group))
            current_group = []
            current_size = 0

    if current_size > 0:
        SC_groups.append(ScaffoldCompareGroup(current_group))

    return SC_groups

def make_scaffoldcomparison_objects(inputs, scaffolds_to_compare, input2scaffolds=None):
    inStrain.logUtils.log_checkpoint("Compare", "CreateScaffoldComparisonObjects", "start")

    # Get the stuff to return
    scaffold2SC = {}

    # Keep this cached; necessary for genome-level operations
    scaffold2length = {}

    # Go through the profiles
    for profile_loc in tqdm(inputs, desc='Loading Profiles into RAM'):
        if not os.path.exists(profile_loc):
            logging.error("IS {0} does not exist! Skipping".format(profile_loc))
            continue

        logging.debug("Loading {0}".format(profile_loc))
        ISP = inStrain.SNVprofile.SNVprofile(profile_loc)
        scaffolds = list(ISP._get_covt_keys())
        name = os.path.basename(ISP.get('bam_loc'))

        if input2scaffolds is not None:
            scaffolds = list(set(scaffolds).intersection(input2scaffolds[profile_loc]))
        if scaffolds_to_compare is not None:
            scaffolds = list(set(scaffolds).intersection(scaffolds_to_compare))

        # Update scaffold to length
        s2l = ISP.get('scaffold2length')
        for scaff, l in s2l.items():
            if scaff in scaffold2length:
                assert l == scaffold2length[scaff]
            else:
                scaffold2length[scaff] = l

        # Create ScaffoldComparison objects
        for scaff in scaffolds:
            if scaff not in scaffold2SC:
                SC = ScaffoldComparison(scaff, scaffold2length[scaff])
                scaffold2SC[scaff] = SC
            scaffold2SC[scaff].add_profile(ISP, name)

    # Figure out which scaffolds to compare
    valid_SCs = [SC for scaff, SC in scaffold2SC.items() if SC.valid()]
    logging.info(f"{len(valid_SCs)} of {len(scaffold2SC.keys())} scaffolds are in at least 2 samples")
    assert len(valid_SCs) > 0, "No scaffolds are shared among the IS objects"

    inStrain.logUtils.log_checkpoint("Compare", "CreateScaffoldComparisonObjects", "end")

    return valid_SCs, scaffold2length

def calc_scaff_sim_matrix(valid_SCs):
    """
    Perform pairwise comparison of scaffolds based on the overlap of profiles
    """
    table = defaultdict(list)
    for i, sc1 in enumerate(valid_SCs):
        for j, sc2 in enumerate(valid_SCs):
            if i <= j:
                continue
            table['sc1'].append(sc1)
            table['sc2'].append(sc2)
            table['similarity'].append(sc1.compare(sc2))
    return pd.DataFrame(table)

def establish_SC_groups(valid_SCs, Cdb, group_length):
    """
    Based on the distance matrix, put valid_SCs into a number of groups with a max of group_length
    """
    # Just do it simple; would be better to use the distance matrix, but oh well
    SC_groups = simple_grouping(valid_SCs, group_length)

    return SC_groups


import logging
import traceback
import pandas as pd
import seaborn as sns

from collections import defaultdict
import numpy as np

import matplotlib

matplotlib.use('Agg')
matplotlib.rcParams['pdf.fonttype'] = 42

import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.plotting.utilities
from inStrain.plotting.utilities import plot_genome
from inStrain.plotting.utilities import estimate_breadth


def mm_plot_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        Mdb = kwargs.get('GWdb', False)
        assert len(Mdb) > 0

        if 'mm' not in Mdb:
            raise Exception('Plot 1 cannot be created when run with --database_mode or --skip_mm_profiling')

        # Add the number of read-pairs
        readLen = int(IS.get_read_length())
        Mdb['read_length'] = readLen
        Mdb['mm'] = Mdb['mm'].astype(int)
        Mdb.loc[:,'ANI_level'] = [(readLen - mm) / readLen for mm in Mdb['mm']]
    except:
        logging.error(
            "Skipping plot 1 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 1")
    name = 'CoverageAndBreadth_vs_readMismatch.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Mdb.groupby('genome'):
        if not plot_genome(genome, IS, **kwargs):
            continue
        mm_plot(mdb, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(6, 4)
        pp.savefig(fig)
        plt.close(fig)

    # Save the figure
    pp.close()
    plt.close('all')

def read_filtering_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        # Prepare
        Mdb = IS.get('mapping_info')
        Mdb = Mdb[Mdb['scaffold'] == 'all_scaffolds']
        Mdb['genome'] = 'all_scaffolds'
        assert len(Mdb) > 0
    except:
        logging.error("Skipping plot 6 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 6")
    name = 'ReadFiltering_plot.pdf'
    pp = PdfPages(plot_dir + name)

    read_filtering_plot(Mdb, title='all scaffolds')

    fig = plt.gcf()
    fig.set_size_inches(6, 4)
    fig.tight_layout()
    pp.savefig(fig)
    plt.close(fig)

    # Save the figure
    pp.close()
    plt.close('all')

def ANI_dist_plot_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        Mdb = prepare_read_ani_dist_plot(IS)
        if len(Mdb['ANI_level'].unique()) == 1:
            raise Exception('Plot 3 cannot be created when run with --database_mode or --skip_mm_profiling')
        assert len(Mdb) > 0
    except:
        logging.error("Skipping plot 3 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 3")
    name = 'readANI_distribution.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Mdb.groupby('genome'):
        if not plot_genome(genome, IS, **kwargs):
            continue
        db = read_ani_dist_plot(mdb, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(6, 4)
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def mm_plot(db, left_val='breadth', right_val='coverage', title='',
            minANI=0.9):
    """
    The input db for this is "mm_genome_info" from "makeGenomeWide" in genomeUtilities.py
    """
    db = db.sort_values('ANI_level')
    sns.set_style('white')

    # breadth
    fig, ax1 = plt.subplots()
    ax1.plot(db['ANI_level'], db[left_val], ls='-', color='blue')
    if left_val == 'breadth':
        ax1.plot(db['ANI_level'], estimate_breadth(db['coverage']), ls='--', color='lightblue')
    ax1.set_ylabel(left_val, color='blue')
    ax1.set_xlabel('Minimum read ANI level')
    ax1.set_ylim(0, 1)

    # coverage
    ax2 = ax1.twinx()
    ax2.plot(db['ANI_level'], db[right_val], ls='-', color='red')
    ax2.set_ylabel(right_val, color='red')
    ax2.set_ylim(0, )

    # asthetics
    plt.xlim(1, max(minANI, db['ANI_level'].min()))

    plt.title(title)

def prepare_read_ani_dist_plot(IS):
    # Make a non-cumulative scaffold table
    covTS = IS.get('covT')
    s2l = IS.get('scaffold2length')
    table = defaultdict(list)
    for scaffold, covT in covTS.items():
        for mm, counts in covT.items():
            lengt = s2l[scaffold]
            counts = counts.append(pd.Series(np.zeros(lengt - len(counts))))

            coverage = np.mean(counts)

            table['scaffold'].append(scaffold)
            table['mm'].append(mm)
            table['coverage'].append(coverage)
            table['length'].append(s2l[scaffold])

    # Make it genome wide
    db = pd.DataFrame(table)
    stb = IS.get('scaffold2bin')
    b2l = IS.get('bin2length')
    gdb = inStrain.genomeUtilities._add_stb(db, stb, verbose=False)

    table = defaultdict(list)
    for mm, Odb in gdb.groupby('mm'):
        for genome, df in Odb.groupby('genome'):
            table['mm'].append(mm)
            table['genome'].append(genome)
            table['length'].append(int(b2l[genome]))

            for col in ['coverage']:
                table[col].append(sum(x * y for x, y in zip(df[col], df['length'])) / b2l[genome])
    db = pd.DataFrame(table)

    # Add the number of read-pairs
    #readLen = int(IS.get('mapping_info')['mean_pair_length'].tolist()[0])
    readLen = int(IS.get_read_length())
    db['read_length'] = readLen
    db['mm'] = db['mm'].astype(int)
    db.loc[:,'read_pairs'] = [int((x*y) / (readLen * 2)) for x, y in zip(db['coverage'], db['length'])]
    db.loc[:,'ANI_level'] = [(readLen - mm)/readLen for mm in db['mm']]

    return db

def read_ani_dist_plot(db, title=None):
    # Plot it
    plt.plot(db['ANI_level'], db['read_pairs'])

    # Adjust x-tickets
    if db['ANI_level'].max() != db['ANI_level'].min():
        plt.gca().set_xlim(db['ANI_level'].max(), db['ANI_level'].min())

    # Other stuff
    rl = int(db['read_length'].tolist()[0])
    plt.xlabel('Read ANI level')
    plt.ylabel("Numbner of read pairs (average length {0}bp)".format(rl))
    plt.title(title)

def read_filtering_plot(db, title=''):
    # Prepare data
    keep_cols = [x for x in db.columns if 'pass' in x] \
                 + ['unfiltered_reads', 'unfiltered_pairs', 'filtered_pairs']
    db = db.melt(id_vars=['genome'], value_vars=keep_cols)

    # Rename
    c2c = {'unfiltered_reads':'Total mapping reads (divided by 2)',
          'unfiltered_pairs':'Total mapped pairs',
          'pass_min_mapq':'Pairs passing mapQ threshold',
          'pass_max_insert':'Pairs passing max insert size threshold',
          'pass_min_insert':'Pairs passing min insert size threshold',
          'pass_filter_cutoff':'Pairs passing ANI threshold',
          'filtered_pairs':'Total filtered pairs'}
    db.loc[:,'variable'] = [c2c[x] if x in c2c else x for x in db['variable']]
    db.loc[:,'value'] = [int(x/2) if y == 'Total mapping reads (divided by 2)' else x for x, y in zip(
                db['value'], db['variable'])]

    # Set up colors
    v2c = {v:'grey' for v in db['variable'].unique()}
    v2c['Total filtered pairs'] = 'green'

    db = db.sort_values(['value', 'variable'], ascending=False)
    ax = sns.barplot(x=db['value'], y=db['variable'], palette=v2c)
    plt.xlabel("Number of read pairs")
    plt.ylabel("")

    # Annotate every single Bar with its value, based on it's width
    offset = db['value'].max() / 12
    total = db[db['variable'] == 'Total mapped pairs']['value'].tolist()[0]
    if total > 0:
        for i, p in enumerate(ax.patches):
            if i == 0:
                continue
            width = p.get_width()
            plt.text(offset + p.get_width(), p.get_y()+0.55*p.get_height(),
                     '{:1.0f}%'.format((width/total)*100),
                     ha='center', va='center')

    # Remove top and right axes
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    plt.title(title)



'''
Opperations related to fasta files
'''

import logging
import pandas as pd
from Bio import SeqIO
from collections import defaultdict

import inStrain.genomeUtilities

def load_fasta(**kwargs):
    '''
    Load the sequences to be profiled. All arguments from the "args" variable

    Return a table listing scaffold name, start, end
    '''
    fasta_loc = kwargs.get('fasta')
    window_length = kwargs.get('window_length')

    scaffolds_to_profile = kwargs.get('scaffolds_to_profile', None)
    use_full_fasta_header = kwargs.get('use_full_fasta_header', False)

    if use_full_fasta_header:
        scaff2sequence = {r.description:r.seq.upper() for r in SeqIO.parse(fasta_loc, "fasta")}
    else:
        scaff2sequence = {r.id:r.seq.upper() for r in SeqIO.parse(fasta_loc, "fasta")}

    # Get scaffold2length
    s2l = {s:len(scaff2sequence[s]) for s in list(scaff2sequence.keys())}

    # Generate splits
    table = defaultdict(list)
    for scaffold, sLen in s2l.items():
        for i, (split_start, split_end) in enumerate(iterate_splits(sLen, window_length)):
            table['scaffold'].append(scaffold)
            table['split_number'].append(i)
            table['start'].append(split_start)
            table['end'].append(split_end)
    Fdb = pd.DataFrame(table)

    # This just takes too long, sadly
    # _validate_splits(Fdb, s2l)

    if scaffolds_to_profile != None:
        Fdb = Fdb[Fdb['scaffold'].isin(scaffolds_to_profile)]
        s2s = {scaff:seq for scaff, seq in scaff2sequence.items() if scaff in scaffolds_to_profile}

    if len(Fdb) == 0:
        logging.error("The provided scaffold list has no overlap with the provided .fasta file!")
        logging.error("Example scaffolds in list: {0}".format("\n".join(scaffolds_to_profile)))
        sys.exit()

    return Fdb, scaff2sequence, s2l # also return s2l - alexcc 5/9/2019: Nah, make it scaff2sequence (s2s) (M.O. 6/10/19)

def iterate_splits(sLen, WINDOW_LEN):
    '''
    Splits are 0-based and double-inclusive
    '''
    numberChunks = sLen // WINDOW_LEN + 1
    chunkLen = int(sLen / numberChunks)

    #print("Scaffold length of {0}, window length of {1}, {2} splits of {3}".format(sLen, WINDOW_LEN, numberChunks, chunkLen))

    start = 0
    end = 0
    for i in range(numberChunks):
        if i + 1 == numberChunks:
            yield start, sLen - 1
        else:
            end += chunkLen
            yield start, end - 1
            start += chunkLen

def _validate_splits(Fdb, s2l):
    '''
    Splits are 0-based and double-inclusive
    '''
    for scaffold, db in Fdb.groupby('scaffold'):
        db['len'] = db['end'] - db['start'] + 1
        if db['len'].sum() != s2l[scaffold]:
            print(db)
        assert db['len'].sum() == s2l[scaffold], [db['len'].sum(), s2l[scaffold]]
        assert db['start'].min() == 0
        assert db['end'].max() == s2l[scaffold] - 1

def filter_fasta(FAdb, s2p, s2l, rl, **kwargs):
    '''
    Filter the .fasta file based on the min number of mapped paired reads
    '''
    min_reads = kwargs.get('min_scaffold_reads', 0)
    min_genome_coverage = kwargs.get('min_genome_coverage', 0)

    if min_genome_coverage > 0:
        FAdb = _filter_genome_coverage(FAdb, s2l, s2p, rl, min_genome_coverage, kwargs.get('stb'))

    if len(FAdb) == 0:
        return FAdb

    # Remove scaffolds without the min number of reads
    FAdb = FAdb[[True if (s2p[s] >= min_reads) else False for s in FAdb['scaffold']]]

    # Sort scaffolds based on the number of reads
    FAdb['filtered_pairs'] = FAdb['scaffold'].map(s2p)
    FAdb = FAdb.sort_values('filtered_pairs', ascending=False)

    return FAdb

def _filter_genome_coverage(FAdb, s2l, s2p, rl, min_genome_coverage, stb_loc):
    '''
    Calcualte the coverage of genomes based on the read filtering, and only keep scaffolds that are above the threshold

    stb_loc should be a list, direct from argument parser
    '''
    cdb = FAdb.drop_duplicates(subset=['scaffold'])
    cdb.loc[:, 'read_pairs'] = cdb['scaffold'].map(s2p)
    cdb.loc[:, 'length'] = cdb['scaffold'].map(s2l)

    stb = inStrain.genomeUtilities.load_scaff2bin(stb_loc)
    cdb = inStrain.genomeUtilities._add_stb(cdb, stb)

    xdb = cdb.groupby('genome')[['read_pairs', 'length']].agg(sum).reset_index()
    xdb['genome_bases'] = xdb['read_pairs'] * rl
    xdb['coverage'] = xdb['genome_bases'] / xdb['length']
    genome_to_rm = set(xdb[xdb['coverage'] < min_genome_coverage]['genome'].tolist())

    scaffolds_to_rm_1 = set(cdb[cdb['genome'].isin(genome_to_rm)]['scaffold'].tolist())
    scaffolds_to_rm_2 = set(cdb[cdb['genome'].isna()]['scaffold'].tolist())
    scaffolds_to_rm = scaffolds_to_rm_1.union(scaffolds_to_rm_2)

    logging.info("{0} of {1} genomes have less than {2}x estimated coverage".format(
            len(genome_to_rm), len(xdb['genome'].unique()), min_genome_coverage))
    logging.info("{0} of the original {1} scaffolds are removed ({2} have a low coverage genome; {3} have no genome)".format(
            len(scaffolds_to_rm), len(cdb['scaffold'].unique()), len(scaffolds_to_rm_1), len(scaffolds_to_rm_2)))

    return FAdb[~FAdb['scaffold'].isin(scaffolds_to_rm)]

def load_scaff_list(list):
    '''
    If this is a text file of scaffolds, return it

    If it's a .fasta file, return a list of scaffolds
    '''

    if list == None:
        return None

    # Try as it its a fasta file
    scaffs = []
    handle = open(list, "r")
    fasta = SeqIO.parse(handle, "fasta")
    for f in fasta:
        scaffs.append(f.id)

    if len(scaffs) > 0:
        handle.close()
        return(set(scaffs))

    else:
        scaffs = []
        handle.close()
        handle = open(list, "r")
        for line in handle.readlines():
            scaffs.append(line.strip())
        handle.close()
        return set(scaffs)



import logging
import traceback
import pandas as pd
import seaborn as sns

from collections import defaultdict
import numpy as np

import matplotlib

matplotlib.use('Agg')
matplotlib.rcParams['pdf.fonttype'] = 42

import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.plotting.utilities
from inStrain.plotting.utilities import plot_genome
from inStrain.plotting.utilities import estimate_breadth

def gene_histogram_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        # Prepare
        db = inStrain.GeneProfile.get_gene_info(IS)
        stb = IS.get('scaffold2bin')
        Gdb = inStrain.genomeUtilities._add_stb(db, stb, verbose=False)
        if 'clonality' in Gdb.columns:
            Gdb.loc[:,'nucl_diversity'] = 1 - Gdb['clonality']
        assert len(Gdb) > 0
    except:
        logging.error("Skipping plot 9 - you don't have all required information. You need to run inStrain profile_genes first")
        if kwargs.get('debug', False):
            traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 9")
    name = 'GeneHistogram_plot.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Gdb.groupby('genome'):
        if not plot_genome(genome, IS, **kwargs):
            continue
        db = gene_histogram_plot(mdb, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(8, 5)
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def gene_histogram_plot(db, title=''):
    COLS = ['coverage', 'nucl_diversity', 'SNPs_per_bp']
    sns.set_style('white')

     # Get set up for multiple rows
    i = len(set(db.columns).intersection(set(COLS)))
    if i > 1:
        fig, ax = plt.subplots(i, 1, sharex=True)
    else:
        ax = {}
        ax[0] = plt.gca()

    i = 0
    for metric in COLS:
        if metric not in set(db.columns):
            continue

        df = db[[metric]].sort_values(metric, ascending=False).reset_index(drop=True).reset_index()
        ax[i].axvline(0, c='black')
        ax[i].axhline(0, c='black')
        ax[i].plot(df['index'], df[metric], marker='o', ms=1)
        ax[i].set_ylabel("{0}".format(metric))
        i += 1

    plt.xlabel('gene index')
    plt.suptitle(title, y=0.999)
    plt.axvline(0)
    #plt.xlim(-1, df['index'].max())


#!/usr/bin/env python

from setuptools import setup, find_packages

from inStrain._version import __version__

setup(name='inStrain',
      version=__version__,
      description='Calculation of strain-level metrics',
      url='https://github.com/MrOlm/inStrain',
      author='Matt Olm and Alex Crits-Christoph',
      author_email='mattolm@berkeley.edu',
      license='MIT',
      package_data={'inStrain': ['helper_files/NullModel.txt']},
      #packages=['inStrain'],
      packages=find_packages(),
      scripts=['bin/inStrain'],
      python_requires='>=3.4.0',
      install_requires=[
          'numpy',
          'pandas>=0.25,!=1.1.3',
          'seaborn',
          'matplotlib',
          'biopython<=1.74',
          'scikit-learn',
          'pytest',
          'tqdm',
          'pysam>=0.15', # This sets a requirement for python 3.7 for now, but so be it. pysam v0.9 (which works on python 3.8) has a broken iterator (has no stop)
          'networkx',
          'h5py',
          'psutil',
          'lmfit',
          'numba'
      ],
      zip_safe=False)



#!/usr/bin/env python

import os
import sys
import json
import h5py
import time
import psutil
import pickle
import logging
import argparse
import traceback
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
import concurrent.futures
from concurrent import futures
from collections import defaultdict

from Bio import SeqIO

if __name__ != '__main__':
    from ._version import __version__

import inStrain.profile.snv_utilities
import inStrain.profile.fasta

import inStrain.SNVprofile
import inStrain.controller
import inStrain.logUtils

C2P = {0:'A', 1:'C', 2:'T', 3:'G'}
def compare_scaffold(scaffold, cur_names, SNPtables, covTs, mLen, null_model, **kwargs):
    '''
    This is the money method thats going to be multithreaded eventually

    Arguments:
        scaffold: name of scaffold (needs to match the SNPtables)
        names: names of the samples; must be in the same order as covTs and SNPtables
        sProfiles: a list of SNVprofiles that correspond to the names list
        mLen: length of the scaffold

    Returns:
        [DataFrame of "scaffold, sample1, sample2, mm, coverage_overlap, ANI",
        DataFrame of SNP locs,
        pair2mm2covOverlap,
        name of scaffold], log
    '''
    log_message = inStrain.logUtils.get_worker_log('Compare', scaffold, 'start')

    # Load arguments
    min_cov = kwargs.get('min_cov', 5)
    min_freq = kwargs.get('min_freq', 5)
    debug = kwargs.get('debug', False)
    fdr = kwargs.get('fdr', 1e-6)
    store_coverage = kwargs.get('store_coverage_overlap', False)
    store_mm_locations = kwargs.get('store_mismatch_locations', False)
    include_self_comparisons = kwargs.get('include_self_comparisons', False)

    # For testing purposes
    if ((scaffold == 'FailureScaffoldHeaderTesting') & (debug)):
        assert False

    if len(cur_names) < 2:
        results = [pd.DataFrame(), pd.DataFrame(), {}, 'skip_{0}'.format(scaffold)]
        log_message += '\n' + inStrain.logUtils.get_worker_log('Compare', scaffold, 'end')
        return (results, log_message)

    # Iterate through pairs
    i = 0
    snpLocs = []
    pair2mm2covOverlap = {}
    table = defaultdict(list)
    for covT1, SNPtable1_ori, name1 in zip(covTs, SNPtables, cur_names):
        i += 1
        j = 0
        for covT2, SNPtable2_ori, name2 in zip(covTs, SNPtables, cur_names):
            j += 1

            if i > j:
                continue

            if (not include_self_comparisons) & (i == j):
                continue

            logging.debug("{2} {0} vs {1} ({3} {4})".format(name1, name2, scaffold,
                        i, j))

            if debug:
                pid = os.getpid()
                process  = psutil.Process(os.getpid())
                bytes_used = process.memory_info().rss
                total_available_bytes = psutil.virtual_memory()
                nm = "\n{4} PID {0} end at {5} with {1} RAM. System has {2} of {3} available".format(
                        pid, bytes_used, total_available_bytes[1], total_available_bytes[0],
                        scaffold, time.time())
                logging.debug(nm)

            mm2overlap, mm2coverage = calc_mm2overlap(covT1, covT2, min_cov=min_cov, verbose=False, debug=debug)
            Mdb = _calc_SNP_count_alternate(SNPtable1_ori, SNPtable2_ori, mm2overlap, null_model, min_freq=min_freq, debug=debug)

            table = _update_overlap_table(table, scaffold, mm2overlap, mm2coverage, Mdb, name1, name2, mLen)

            if store_mm_locations:
                Mdb['name1'] = name1
                Mdb['name2'] = name2
                Mdb['scaffold'] = scaffold
                snpLocs.append(Mdb)

            if store_coverage:
                pair2mm2covOverlap['-vs-'.join(sorted([name1, name2]))] = mm2overlap

    # logging.debug("Returning {0} {1} {2}".format(scaffold, i, j))

    Cdb = pd.DataFrame(table)
    # if len(Cdb) > 0:
    #     Cdb['coverage_overlap'] = Cdb['coverage_overlap'].astype('int64')

    if len(snpLocs) > 0:
        Mdb = pd.concat(snpLocs, sort=False)
    else:
        Mdb = pd.DataFrame()

    results = [Cdb, Mdb, pair2mm2covOverlap, scaffold]
    log_message += '\n' + inStrain.logUtils.get_worker_log('Compare', scaffold, 'end')
    return (results, log_message)

def calc_mm2overlap(covT1, covT2, min_cov=5, verbose=False, debug=False):
    '''
    Calculate mm2overlap for a pair of covTs

    Coverage is calculated as cov = len(coveredInBoth) / len(coveredInEither)
    This means that its the percentage of bases that are covered by both

    Returns:
        mm2overlap -> dictionary to array of "True" where there's overlap and "False" where both are compared but there's not overlap
        mm2coverage -> dictionary of mm -> the alignment coverage
    '''
    mm2overlap = {}
    mm2coverage = {}

    # if debug != False:
    #     scaffold, name1, name2 = debug

    mms = sorted(list(set(covT1.keys()).union(set(covT2.keys()))))
    cov1 = pd.Series()
    cov2 = pd.Series()
    for mm in mms:
        if mm in covT1:
            cov1 = cov1.add(covT1[mm], fill_value=0)
        if mm in covT2:
            cov2 = cov2.add(covT2[mm], fill_value=0)

        # Figure out where each has min coverage
        T1 = set(cov1[(cov1 >= min_cov)].index)
        T2 = set(cov2[(cov2 >= min_cov)].index)

        # Figure out the total possible overlap
        coveredInEither = T1.union(T2)

        # Figure out where there's overlap in both
        coveredInBoth = T1.intersection(T2)

        # Calculate coverage
        if len(coveredInEither) > 0:
            cov = len(coveredInBoth) / len(coveredInEither)
        else:
            cov = 0

        # Save
        mm2overlap[mm] = coveredInBoth
        mm2coverage[mm] = cov

    return mm2overlap, mm2coverage

def _gen_blank_Mdb(COLUMNS):
    '''
    COLUMNS = ['position', 'consensus_SNP', 'population_SNP', 'mm'
    'con_base_1', 'ref_base_1', 'var_base_1', 'position_coverage_1',
    'A_1', 'C_1', 'T_1', 'G_1',
    'con_base_2', 'ref_base_2', 'var_base_2', 'position_coverage_2',
    'A_2', 'C_2', 'T_2', 'G_2']
    '''

    return pd.DataFrame({c:[] for c in COLUMNS})


def _calc_SNP_count_alternate(SNPtable1, SNPtable2, mm2overlap, null_model, min_freq=.05, debug=False):

    mm2ANI = {}
    mm2popANI = {}
    dbs = []

    # Get the null model for SNP calling
    model_to_use = null_model

    # Constant for the SNP dataframe
    SNP_COLUMNS = ['position', 'con_base', 'ref_base', 'var_base',
    'position_coverage', 'A', 'C', 'T', 'G']

    # Constant for the output dataframe
    OUT_COLUMNS = ['position', 'consensus_SNP', 'population_SNP', 'mm',
    'con_base_1', 'ref_base_1', 'var_base_1', 'position_coverage_1',
    'A_1', 'C_1', 'T_1', 'G_1',
    'con_base_2', 'ref_base_2', 'var_base_2', 'position_coverage_2',
    'A_2', 'C_2', 'T_2', 'G_2']

    RENAME_COLUMNS = ['con_base', 'ref_base', 'var_base', 'position_coverage',
                    'A', 'C', 'T', 'G']

    # Iterate mm levels
    for mm, cov_arr in mm2overlap.items():

        # Subset to bases that have coverage in both
        covs = set(cov_arr)

        # These represent relevant counts at these posisions
        if len(SNPtable1) > 0:
            s1_all = SNPtable1[[(p in covs) for p in SNPtable1['position'].values]].drop_duplicates(
                        subset=['position'], keep='last')
            del s1_all['mm']
            if len(s1_all) == 0:
                s1_all = None
        else:
            #s1_all = _gen_blank_SNPdb(SNP_COLUMNS)
            s1_all = None

        if len(SNPtable2) > 0:
            s2_all = SNPtable2[[(p in covs) for p in SNPtable2['position'].values]].drop_duplicates(
                        subset=['position'], keep='last')
            del s2_all['mm']
            if len(s2_all) == 0:
                s2_all = None
        else:
            #s2_all = _gen_blank_SNPdb(SNP_COLUMNS)
            s2_all = None

        # Merge
        if (s1_all is None) & (s2_all is None):
            #Mdb = _gen_blank_Mdb(OUT_COLUMNS)
            Mdb = None

        elif s1_all is None:
            Mdb = s2_all.rename(columns={c:c + '_2' for c in RENAME_COLUMNS})
            for c in RENAME_COLUMNS:
                Mdb[c + '_1'] = np.nan

        elif s2_all is None:
            Mdb = s1_all.rename(columns={c:c + '_1' for c in RENAME_COLUMNS})
            for c in RENAME_COLUMNS:
                Mdb[c + '_2'] = np.nan

        else:
            Mdb = pd.merge(s1_all, s2_all, on='position', suffixes=('_1', '_2'), how='outer', copy=False)

        if Mdb is not None:
            Mdb.loc[:,'consensus_SNP'] = Mdb.apply(call_con_snps, axis=1)
            Mdb.loc[:,'population_SNP'] = Mdb.apply(call_pop_snps, axis=1, args=(model_to_use, min_freq))

            Mdb['mm'] = mm
            Mdb = Mdb[OUT_COLUMNS]

            # Only keep SNPs
            Mdb = Mdb[Mdb['consensus_SNP'] | Mdb['population_SNP'] ]

            dbs.append(Mdb)

    if len(dbs) > 0:
        Mdb = pd.concat(dbs, sort=False)
    else:
        Mdb = _gen_blank_Mdb(OUT_COLUMNS)
    return Mdb

def call_con_snps(row):
    '''
    Call a SNP if the consensus sequnces aren't the same
    '''

    # This was only a SNP in the first sapmle
    if row['con_base_1'] != row['con_base_1']:
        return row['con_base_2'] != row['ref_base_2']

    # This was only a SNP in the second sapmle
    if row['con_base_2'] != row['con_base_2']:
        return row['con_base_1'] != row['ref_base_1']

    # This is a SNP in both samples
    return row['con_base_1'] != row['con_base_2']

def is_present(counts, total, model, min_freq):
    '''
    Return true if the base counts represented by "counts" are detected above background
    '''
    if total in model:
        min_bases = model[total]
    else:
        min_bases = model[-1]

    return (counts >= min_bases) and ((float(counts) / total) >= min_freq)

def call_pop_snps(row, model, min_freq):
    '''
    To be applied to a DataFrame

    Call a SNP if you can't find the consenus of 1 in 2 AND
    you can't find the consensus of 2 in 1 AND
    1 and 2 don't share a minor allele
    '''
    # Are the consensus bases the same?
    if row['con_base_1'] == row['con_base_2']:
        return False

    # Is it a SNP in only one? If so, see if the reference is still there
    if (row['con_base_1'] != row['con_base_1']) | (row['con_base_2'] != row['con_base_2']):

        # In this case, is consensus allele still detected?
        if (row['con_base_1'] != row['con_base_1']):
            count = row['{0}_2'.format(row['ref_base_2'])]
            total = row['position_coverage_2']
            if is_present(count, total, model, min_freq):
                return False

        elif (row['con_base_2'] != row['con_base_2']):
            count = row['{0}_1'.format(row['ref_base_1'])]
            total = row['position_coverage_1']
            if is_present(count, total, model, min_freq):
                return False

        return True

    ### OK, so it's a SNP in both ###

    # Look for con_base_1 in sample 2
    try:
        count = row["{0}_2".format(row['con_base_1'])]
        total = row['position_coverage_2']
        if is_present(count, total, model, min_freq):
            return False
    except:
        print(row)

    # Look for con_base_2 in sample 1
    count = row["{0}_1".format(row['con_base_2'])]
    total = row['position_coverage_1']
    if is_present(count, total, model, min_freq):
        return False

    # Look for minor in both samples
    if 'allele_count_1' in row:
        if (row['allele_count_1'] > 1) & (row['allele_count_2'] > 1):
            if row['var_base_1'] == row['var_base_2']:
                return False

    elif 'morphia_1' in row:
        if (row['morphia_1'] > 1) & (row['morphia_2'] > 1):
            if row['var_base_1'] == row['var_base_2']:
                return False

    return True

def _is_snp(db1, db2, position, p2c1, p2c2, min_freq, model_to_use, debug=False, compare_consensus_bases=False):
    '''
    Determine if the consensus base of db1 is a SNP in db2 and vice versa
    '''
    # Try and out quick
    if p2c1[position] == p2c2[position]:
        return False
    if compare_consensus_bases:
        return True

    # otime = time.time()
    # print("checkpoint {0} {1}".format(0, time.time() - otime))
    # These are sorted by mm above
    dd1 = db1[db1['position'] == position]#.sort_values('mm', ascending=False)#.drop_duplicates(subset='position', keep='last')
    dd2 = db2[db2['position'] == position]#.sort_values('mm', ascending=False)#.drop_duplicates(subset='position', keep='last')

    #print("checkpoint {0} {1}".format(1, time.time() - otime))

    assert len(dd1) > 0, [position, 'one']
    assert len(dd2) > 0, [position, 'two']

    #print("checkpoint {0} {1}".format(2, time.time() - otime))

    con1 = p2c1[position]
    con2 = p2c2[position]

    #print("{0} in {1} is {2}".format(position, p2c1, con1))

    #print("checkpoint {0} {1}".format(3, time.time() - otime))

    # Check if the consensus of db1 is a SNP in db2
    #counts2 = dd2.iloc[-1][C2P[con1]]
    counts2 = dd2.iloc[-1][con1]
    total = dd2.iloc[-1]['position_coverage']

    if total in model_to_use:
        min_bases = model_to_use[total]
    else:
        min_bases = model_to_use[-1]

    if (counts2 >= min_bases) and ((float(counts2) / total) >= min_freq):
        pass
    else:
        return True

    # Check the opposite
    #counts1 = dd1.iloc[-1][C2P[con2]]
    counts1 = dd1.iloc[-1][con2]
    total = dd1.iloc[-1]['position_coverage']

    if total in model_to_use:
        min_bases = model_to_use[total]
    else:
        min_bases = model_to_use[-1]

    if (counts1 >= min_bases) and ((float(counts1) / total) >= min_freq):
        pass
    else:
        return True

    # Looks like you're not a SNP, then!
    if debug:  print('exit2')

    #print("checkpoint {0} {1}".format(5, time.time() - otime))

    return False

def _update_overlap_table(table, scaffold, mm2overlap, mm2coverage, Mdb, name1, name2, mLen):
    '''
    covarage_overlap = the percentage of bases that are either covered or not covered in both
        - So if both scaffolds have 0 coverage, this will be 1
    percent_genome_compared = the percentage of bases in the scaffolds that are covered by both
        - So if both scaffolds have 0 coverave, this will be 0
    compared_bases_count = the number of considered bases
    '''
    rel_mms = set(list(mm2overlap.keys()))
    got_mms = set()

    for mm, mdb in Mdb.groupby('mm'):
        if mm not in rel_mms:
            continue
        got_mms.add(mm)

        overlap = mm2overlap[mm]
        bases = len(overlap)

        table['mm'].append(mm)
        table['scaffold'].append(scaffold)
        table['name1'].append(name1)
        table['name2'].append(name2)
        table['coverage_overlap'].append(mm2coverage[mm])
        table['compared_bases_count'].append(bases)
        table['percent_genome_compared'].append(bases/mLen)
        table['length'].append(mLen)

        snps = len(mdb[mdb['consensus_SNP'] == True])
        popsnps = len(mdb[mdb['population_SNP'] == True])

        table['consensus_SNPs'].append(snps)
        table['population_SNPs'].append(popsnps)

        if bases == 0:
            table['popANI'].append(np.nan)
            table['conANI'].append(np.nan)
        else:
            table['conANI'].append((bases - snps) / bases)
            table['popANI'].append((bases - popsnps) / bases)

    # Doing this to allow the group-by
    for mm in rel_mms - got_mms:
        overlap = mm2overlap[mm]
        bases = len(overlap)

        table['mm'].append(mm)
        table['scaffold'].append(scaffold)
        table['name1'].append(name1)
        table['name2'].append(name2)
        table['coverage_overlap'].append(mm2coverage[mm])
        table['compared_bases_count'].append(bases)
        table['percent_genome_compared'].append(bases / mLen)
        table['length'].append(mLen)

        snps = 0
        popsnps = 0

        table['consensus_SNPs'].append(snps)
        table['population_SNPs'].append(popsnps)

        if bases == 0:
            table['popANI'].append(np.nan)
            table['conANI'].append(np.nan)
        else:
            table['conANI'].append((bases - snps) / bases)
            table['popANI'].append((bases - popsnps) / bases)

    return table


'''
This contains a number of methods and functions to calculate the iRep metric

https://github.com/christophertbrown/iRep

https://www.nature.com/articles/nbt.3704
'''

import os
import sys
import glob
import scipy
import lmfit
import scipy.signal
import numpy as np
import pandas as pd
import seaborn as sns

from Bio import SeqIO
from collections import defaultdict

def calculate_iRep_from_coverage_array(rcov, num_contigs, gc_windows=None):
    '''
    Argumemts:
        rcov: genome-wide array of coverage values
        num_contigs: number of contigs in genome
        gc_windows: GC content of windows to do GC coverage correction

    Returns:
        iRep: iRep value or np.nan if filtered out
        iRep_junk: Dictionary of other iRep outputs
    '''
    FilterChriteria = {'kept_windows':np.nan, # Calculated in coverage_windows
                       'avg_cov':np.nan, # Calculated on raw array
                       'r2':np.nan, # Calculated from iRep itself
                       'fragMbp':np.nan } # Calculated from iRep itself

    # Filter chriteria
    length = len(rcov)
    FilterChriteria['avg_cov'] = np.mean(rcov)
    FilterChriteria['fragMbp'] = num_contigs/(float(length)/1000000)

    # Calculate the windows
    oIdb = _iRep_windows(rcov)

    # Add GC if appropriate
    if gc_windows is not None:
        oIdb = pd.merge(oIdb, gc_windows, on='index')

    # Filter out junk windows
    Idb = _iRep_filter_windows(oIdb, on='coverage')
    FilterChriteria['kept_windows'] = len(Idb) / len(oIdb)

    # Get raw iRep values
    Idb.loc[:,'coverage_OLT'] = _iRep_log_transform(Idb['coverage'])
    iRep = _calc_iRep(Idb, length, on='coverage_OLT', FilterChriteria=FilterChriteria)
    FilterChriteria['unfiltered_raw_iRep'] = iRep

    # Get gc-corrected iRep values
    FilterChriteria['iRep_GC_corrected'] = False
    if gc_windows is not None:
        Idb = _iRep_gc_bias(Idb)
        Idb.loc[:,'coverage_LT'] = _iRep_log_transform(Idb['corrected_coverage'])
        iRep = _calc_iRep(Idb, length, on='coverage_LT')
        FilterChriteria['unfiltered_iRep'] = iRep
        FilterChriteria['iRep_GC_corrected'] = True

    # Get raw iRep values
    Idb.loc[:,'coverage_OLT'] = _iRep_log_transform(Idb['coverage'])
    iRep = _calc_iRep(Idb, length, on='coverage_OLT', FilterChriteria=FilterChriteria)
    FilterChriteria['unfiltered_raw_iRep'] = iRep

    # See if iRep passes
    if (FilterChriteria['kept_windows'] < 0.98) or \
        (FilterChriteria['avg_cov'] < 5) or \
        (FilterChriteria['r2'] < 0.9) or \
        (FilterChriteria['fragMbp'] > 175):

        iRep = np.nan

    return iRep, FilterChriteria

def generate_gc_windows(order, scaff2sequence, mask_edges=100):
    '''
    Calculate the GC content for windows across a .fasta file

    Argumnets:
        order = list of scaffolds in the order they should be in
        scaff2sequence = scaffold2sequence (scaff2sequence = SeqIO.to_dict(SeqIO.parse(fasta_loc, "fasta")))
        mask_edges = remove this many bp from start and end of contigs

    Modified for speed
    '''
    # Load the .fasta
    #scaff2sequence = SeqIO.to_dict(SeqIO.parse(file_loc, "fasta"))

    # Make the genome sequence into a list
    splits = []
    for scaff in order:
        seq = scaff2sequence[scaff]
        if mask_edges:
            seq = seq[100:len(seq)-100]
            splits.append(seq)
            #genome_seq.extend(seq)
    genome_seq = sum(splits, [])

    # Calculate GC content
    gcdb = _iRep_gc_content(genome_seq)

    return gcdb

def _iRep_gc_content(seq, window = 5000, slide = 100):
    """
    iRep gc_content message

    calculate gc content over sequence windows
    """
    # convert GC
    replacements = {'G':1, 'C':1, 'A':0, 'T':0, 'N':0}
    GC = [] # G - C
    for base in seq:
        try:
            GC.append(replacements[base.upper()])
        except:
            GC.append(0)
    # calculate gc content over sliding windows
    i = 0
    weights = np.ones(window)
    table = defaultdict(list)
    for gc in scipy.signal.fftconvolve(GC, weights, 'valid').tolist()[0::slide]:
        table['index'].append(i)
        table['GC_content'].append(gc/window)
        i += slide
    return pd.DataFrame(table)

def _iRep_windows(cov, window=5000, slide=100, FilterChriteria=None):
    '''
    Replicates *coverage_windows()*
    '''
    table = defaultdict(list)

    i = 0
    weights = np.ones(window)
    for c in scipy.signal.fftconvolve(cov, weights, 'valid').tolist()[0::slide]:
        table['index'].append(i)
        table['coverage'].append(c/window)
        i += slide

    return pd.DataFrame(table)

def _iRep_filter_windows(cov, on='coverage', mdif = float(8)):
    '''
    Replicates *filter_windows()*

    Remove windows with weird coverage

    Edited to improve speed in version 1.3.0l
    '''
    med = np.median(cov[on])

    return cov[[True if \
            ((y > 0) and (med > 0) and
            (abs(float(max([y, med])) / float(min([y, med]))) <= mdif))
            else False for y in cov[on]]]
    # keeper_index = []
    # for i, row in cov.iterrows():
    #     y = row[on]
    #     if y <= 0 or med <= 0:
    #         continue
    #     if abs(float(max([y, med])) / float(min([y, med]))) > mdif:
    #         continue
    #     keeper_index.append(row['index'])
    #
    # return cov[cov['index'].isin(keeper_index)]

def _iRep_log_transform(array):
    lt = []
    eps = 1e-50
    for i in array:
        if i < eps:
            lt.append(np.log2(eps))
        else:
            lt.append(np.log2(i))
    return lt

def _calc_iRep(db, length, on='coverage_OLT', FilterChriteria=None):
    '''
    Replicates iRep_from_windows
    '''
    Ys = sorted(list(db[on]))
    windows = len(Ys)

    if windows == 0:
        return np.nan

    dif = float(length)/float(windows)
    Xs = [int(i * dif) + 1 for i, value in enumerate(Ys, 0)]
    Xt, Yt = trim_data((Xs, Ys), xy = True)
    db = pd.DataFrame({'index':Xt, 'cov':Yt})
    m, b, fit, r2, info = fit_coverage((Xt, Yt, None, True))
    iRep = 2**(m * length)

    if FilterChriteria is not None:
        FilterChriteria['r2'] = r2

    return iRep

def trim_data(data, xy, p = 0.1):
    """
    RIGHT FROM iREP

    remove data from ends of sorted list
    """
    if xy is False:
        length = len(data)
        num = int(length * (p/2))
        return data[num:length - num]
    X, Y = data
    length = len(X)
    num = int(length * (p/2))
    return X[num:length - num], Y[num:length - num]

def fit_coverage(pars):
    """
    RIGHT FROM iREP

    fit line to sorted coverage values to get slope
    """
    x, y, info, return_fit = pars
    if len(x) <= 2: # make sure there is more data than parameters
        if return_fit is False:
            return (False, False, False, info)
        else:
            return (False, False, False, False, info)
    # Parameters
    Pars = lmfit.Parameters()
    ## y = mx + b
    Pars.add('m', value = 1, vary = True)
    Pars.add('b', value = 1)
    # fit data to model
    mi = lmfit.minimize(coverage_function, Pars, args = (x,), \
            kws = {'data':y, 'printPs':False}, method = 'leastsq')
    # calculate r-squared
    r2 = 1 - (mi.residual.var() / np.var(y))
    if return_fit is False:
        return (mi.params['m'].value, mi.params['b'].value, r2, info)
    # get fitted values
    fit = [x, coverage_function(mi.params, x)]
    return (mi.params['m'].value, mi.params['b'].value, fit, r2, info)

def coverage_function(pars, X, data = None, printPs = False):
    """
    RIGHT FROM iREP

    linear function for sorted coverage profile
    y = mx + b
    """
    m = pars['m'].value
    b = pars['b'].value
    if printPs is True:
        print('m: %s b: %s' % \
            ('{:,}'.format(int(m)), '{:,}'.format(int(b))))
    results = [float(m * x) + b for x in X]
    if data is None:
        return np.asarray(results)
    return np.asarray([y - data[i] for i, y in enumerate(results)]) # model - data

def _iRep_gc_bias(Idb, correction_threshold=0.0):
    '''
    iRep gc_bias method
    '''
    # fit line
    m, b, fit, r2, l = fit_coverage((Idb['GC_content'].tolist(), Idb['coverage'].tolist(), False, True))

    # remove outliers
    Idb.loc[:,'error'] = [abs(cov - (m * gc + b)) for gc, cov in zip(Idb['GC_content'], Idb['coverage'])]
    try:
        cutoff = sorted(Idb['error'].tolist(), reverse = True)[int(len(Idb['error'])*0.01)]
    except:
        cutoff = 0
    FIdb = Idb[~(Idb['error'] >= cutoff)]

    # re-fit with filtered data
    m, b, fit, r2, l = fit_coverage((FIdb['GC_content'].tolist(), FIdb['coverage'].tolist(), False, True))
    if r2 < correction_threshold:
        Idb['corrected_coverage'] = Idb['coverage']
        return Idb

    # correct coverge
    corrected = [[], []]
    av = np.average(Idb['coverage'])
    Idb['corrected_coverage'] = [cov + (av - (m * gc + b)) for cov, gc in zip(Idb['coverage'], Idb['GC_content'])]

    return Idb



"""
This handles the argument parsing and logic for profiling bam files
"""

# Import packages
import os
import copy
import logging
import pandas as pd
from tqdm import tqdm
import multiprocessing

import inStrain.logUtils
import inStrain.SNVprofile
import inStrain.GeneProfile

import inStrain.profile.linkage
import inStrain.profile.profile_utilities
import inStrain.profile.snv_utilities


class BamProfileController(object):
    """
    Handle the logic of profiling
    """

    def __init__(self, bam, Fdb, sR2M, ISP_loc, **kwargs):
        """
        The requirements are bam, Fdb, and sR2M

        Arguments:
            bam = location of .bam file
            Fdb = dictionary listing fasta locations to profile
            sR2M = dictionary of scaffold -> read pair -> number of mm
            ISP_loc = location of ISP profile to store results
        """
        self.bam_loc = bam
        self.sR2M = sR2M
        self.Fdb = Fdb
        self.ISP_loc = ISP_loc
        self.kwargs = kwargs

    def main(self):
        """
        Profile the bam with inStrain using the options in kwargs
        """
        # Get arguments
        self.gen_prof_args()

        # Generate commands
        self.make_profile_commands()

        # Run the multiprocessing
        self.run_profile_processing()

        # Return the results as an SNVprofile
        return self.ISP

    def gen_prof_args(self):
        """
        Generate a set of arguments to be passed to prepare_comands
        """
        # Make a special copy of the arguments
        profArgs = copy.deepcopy(self.kwargs)
        scaff2sequence = profArgs.pop('s2s')
        self.scaff2sequence = scaff2sequence
        self.profArgs = profArgs
        self.scaffold_num = len(self.Fdb['scaffold'].unique())

        # Generate the null model
        fdr = profArgs.get('fdr', 1e-6)
        null_loc = os.path.dirname(__file__) + '/../helper_files/NullModel.txt'
        null_model = inStrain.profile.snv_utilities.generate_snp_model(null_loc, fdr=fdr)
        self.null_model = null_model

        # Parse for ProfileGenes as needed
        if self.kwargs.get('gene_file') is not None:
            self.gen_genes_args()
        else:
            self.profile_genes = False

    def gen_genes_args(self):
        """
        Parse arguments as needed to profile genes during the "merge" step
        """
        gene_file = self.kwargs.get('gene_file')

        inStrain.logUtils.log_checkpoint('Profile', 'Loading_genes', 'start')
        scaff2geneinfo, scaff2gene2sequence = inStrain.GeneProfile.parse_genes(gene_file, **self.kwargs)

        # Filter gene database to relevant scaffolds
        scaffs = set(self.scaff2sequence.keys())
        for scaff in scaffs:
            if scaff not in scaffs:
                del scaff2geneinfo[scaff]

        self.profile_genes = True
        self.scaff_2_gene_database = scaff2geneinfo
        self.scaff2gene2sequence = scaff2gene2sequence
        inStrain.logUtils.log_checkpoint('Profile', 'Loading_genes', 'end')

    def make_profile_commands(self):
        """
        Make a set of commands to run with profile
        """
        logging.debug('Creating commands')

        cmd_groups, Sprofile_dict, s2splits = prepare_commands(self.Fdb, self.bam_loc,
                                                               self.scaff2sequence, self.profArgs, self.sR2M,
                                                               self)
        self.cmd_groups = cmd_groups
        self.Sprofile_dict = Sprofile_dict
        self.s2splits = s2splits

        logging.debug('There are {0} cmd groups'.format(len(cmd_groups)))

    def run_profile_processing(self):
        """
        Do the actual multiprocessing involved
        """
        # Establish command and result queues
        self.make_profile_queues()

        # Spawn worker processes for profiling splits
        self.spawn_profile_workers()

        # Get results from profiling splits
        self.recieve_profile_results()

        # Spwan worker processes for merging splits
        self.spawn_profile_merge_workers()

        # Get results from merging splits
        self.recieve_merge_results()

        # Collate and return results
        ISP = inStrain.profile.profile_utilities.gen_snv_profile(
            [s for s in self.Sprofiles if s is not None],
            ISP_loc=self.ISP_loc,
            **self.kwargs)

        # Store gene stuff if you have it
        if self.profile_genes:
            ISP.store('genes_fileloc', self.kwargs.get('gene_file'), 'value',
                      'Location of genes file that was used to call genes')

            gdb = pd.concat([x for x in self.scaff_2_gene_database.values()])
            ISP.store('genes_table', gdb, 'pandas',
                      'Location of genes in the associated genes_file')

            if self.kwargs.get('store_everything', False):
                ISP.store('scaff2gene2sequence', self.scaff2gene2sequence, 'pickle',
                          'Dicitonary of scaffold -> gene -> nucleotide sequence')

        self.ISP = ISP

    def make_profile_queues(self):
        """
        Make the queues that will be used for profile multiprocessing
        """
        inStrain.logUtils.log_checkpoint("Profile", "initialize_multiprocessing", "start")

        # Get the arguments
        Sprofile_dict = self.Sprofile_dict
        cmd_groups = self.cmd_groups


        # Create the queues
        ctx = multiprocessing.get_context('spawn')
        self.ctx = ctx

        manager = self.ctx.Manager()
        self.Sprofile_dict = manager.dict(Sprofile_dict)  # Holds a synced directory of splits

        self.log_list = ctx.Queue()
        self.split_cmd_queue = ctx.Queue()
        self.sprofile_cmd_queue = ctx.Queue()
        self.sprofile_result_queue = ctx.Queue()
        self.Sprofiles = []

        # Submit commands to the queues
        logging.debug("Submitting split tasks to queue")
        for i, cmd_group in enumerate(cmd_groups):
            self.split_cmd_queue.put(cmd_group)

        logging.debug("Submitting merge tasks to queue")
        c = 0
        for cmd_group in self.gen_merge_cmds():
            self.sprofile_cmd_queue.put(cmd_group)
            c += 1
        self.num_merge_groups = c

        inStrain.logUtils.log_checkpoint("Profile", "initialize_multiprocessing", "end")

    def gen_merge_cmds(self):
        processes = self.profArgs.get('processes', 6)
        s2splits = self.s2splits
        scaffolds_to_profile = list(s2splits.keys())

        s2g = {}
        if self.profile_genes:
            for scaff, gdb in self.scaff_2_gene_database.items():
                s2g[scaff] = len(gdb)
        for scaff, splits in s2splits.items():
            if scaff not in s2g:
                s2g[scaff] = splits * 50


        SECONDS = min(60, sum(inStrain.GeneProfile.calc_estimated_runtime(s2g[scaffold]) for scaffold in scaffolds_to_profile) / (
                    processes + 1))

        cmds = []
        seconds = 0
        for scaff, splits in s2splits.items():
            Sprofile = inStrain.profile.profile_utilities.ScaffoldSplitObject(splits)
            Sprofile.scaffold = scaff
            Sprofile.profile_genes = self.profile_genes

            if self.profile_genes:
                self.add_gene_info(Sprofile)

            # Add estimated seconds
            seconds += inStrain.GeneProfile.calc_estimated_runtime(s2g[scaff])
            cmds.append(Sprofile)

            # See if you're done
            if seconds >= SECONDS:
                yield cmds
                seconds = 0
                cmds = []

        if len(cmds) > 0:
            yield cmds

    def add_gene_info(self, Sprofile):
        if Sprofile.scaffold in self.scaff_2_gene_database:
            Sprofile.gene_database = self.scaff_2_gene_database[Sprofile.scaffold]
            Sprofile.gene2sequence = self.scaff2gene2sequence[Sprofile.scaffold]
        else:
            return


    def spawn_profile_workers(self):
        """
        Spawn worker threads for profiling splits; or just run them all if a single core
        """
        p = int(self.kwargs.get('processes', 6))

        if p > 1:
            inStrain.logUtils.log_checkpoint("Profile", "SpawningSplitWorkers", "start")
            self.processes = []
            for i in range(0, p):
                self.processes.append(self.ctx.Process(
                    target=inStrain.profile.profile_utilities.split_profile_worker,
                    args=(self.split_cmd_queue,
                          self.Sprofile_dict,
                          self.log_list,
                          self.null_model,
                          self.bam_loc)))
            for proc in self.processes:
                proc.start()
            inStrain.logUtils.log_checkpoint("Profile", "SpawningSplitWorkers", "end")

        else:
            inStrain.profile.profile_utilities.split_profile_worker(self.split_cmd_queue,
                                                                    self.Sprofile_dict,
                                                                    self.log_list,
                                                                    self.null_model,
                                                                    self.bam_loc,
                                                                    single_thread=True)
            logging.info("Done profiling splits")

    def recieve_profile_results(self):
        """
        Get the results from the queues for profiling splits
        """
        p = int(self.kwargs.get('processes', 6))
        total_cmd_count = len(self.cmd_groups)

        if p > 1:
            # Set up progress bar
            pbar = tqdm(desc='Profiling splits: ', total=total_cmd_count)

            # Get the splits
            received_splits = 0
            while received_splits < total_cmd_count:
                try:
                    log = self.log_list.get()
                    logging.debug(log)
                    pbar.update(1)
                    received_splits += 1
                except KeyboardInterrupt:
                    for proc in self.processes:
                        proc.terminate()
                    break

            # Close progress bar
            pbar.close()
            inStrain.logUtils.log_checkpoint("Profile", "TerminatingSplitWorkers", "start")
            for proc in self.processes:
                proc.terminate()
            inStrain.logUtils.log_checkpoint("Profile", "TerminatingSplitWorkers", "end")

        else:
            # Get the splits
            received_splits = 0
            while received_splits < total_cmd_count:
                try:
                    log = self.log_list.get(timeout=5)
                    logging.debug(log)
                    received_splits += 1
                except:
                    logging.warning("Missing splits; {0} {1}".format(self.cmd_groups, self.split_cmd_queue))
                    assert False

    def spawn_profile_merge_workers(self):
        """
        Spawn worker threads for merging splits; or just run them all if a single core
        """
        p = int(self.kwargs.get('processes', 6))

        logging.debug('Establishing processes for merging')
        if p > 1:
            self.processes = []
            for i in range(0, p):
                self.processes.append(self.ctx.Process(
                    target=inStrain.profile.profile_utilities.merge_profile_worker,
                    args=(self.sprofile_cmd_queue,
                          self.Sprofile_dict,
                          self.sprofile_result_queue,
                          self.null_model)))
            for proc in self.processes:
                proc.start()

        else:
            inStrain.profile.profile_utilities.merge_profile_worker(
                self.sprofile_cmd_queue,
                self.Sprofile_dict,
                self.sprofile_result_queue,
                self.null_model,
                single_thread=True)

    def recieve_merge_results(self):
        """
        Get the results from queues after doing merges
        """
        p = int(self.kwargs.get('processes', 6))

        Sprofiles = []
        if p > 1:
            # Set up progress bar
            pbar = tqdm(desc='Merging splits and profiling genes: ', total=self.num_merge_groups)

            # Get results
            received_profiles = 0
            while received_profiles < self.num_merge_groups:
                try:
                    Sprofile_group = self.sprofile_result_queue.get()
                    for Sprofile in Sprofile_group:
                        if Sprofile is not None:
                            logging.debug(Sprofile.merge_log)
                            Sprofiles.append(Sprofile)
                    pbar.update(1)
                    received_profiles += 1
                except KeyboardInterrupt:
                    break

            # Close multi-processing
            for proc in self.processes:
                proc.terminate()

            # Close progress bar
            pbar.close()

        else:
            # Get results
            received_profiles = 0
            while received_profiles < self.num_merge_groups:
                Sprofile_group = self.sprofile_result_queue.get(timeout=5)
                for Sprofile in Sprofile_group:
                    logging.debug(Sprofile.merge_log)
                    Sprofiles.append(Sprofile)
                received_profiles += 1

        self.Sprofiles = Sprofiles


class profile_command():
    """
    This is a stupid object that just holds the arguments to profile a split
    """

    def __init__(self):
        pass


def prepare_commands(Fdb, bam, scaff2sequence, args, sR2M, Sprofile):
    """
    Make and iterate profiling commands
    Doing it in this way makes it use way less RAM
    """

    # Make the SplitProfile command groups
    processes = args.get('processes', 6)
    s2p = args.get('s2p', None)
    if s2p is not None:
        SECONDS = min(60,
                      sum(calc_estimated_runtime(s2p[scaff]) for scaff in Fdb['scaffold'].unique()) / (processes + 1))
    else:
        SECONDS = 60

    cmd_groups = []
    Sdict = {}
    s2splits = {}

    seconds = 0
    cmds = []
    for scaff, db in Fdb.groupby('scaffold'):
        s2splits[scaff] = len(db)

        for i, row in db.iterrows():

            # make this command
            start = int(row['start'])
            end = int(row['end'])

            cmd = profile_command()
            cmd.scaffold = scaff
            cmd.R2M = sR2M[scaff]
            cmd.samfile = bam
            cmd.arguments = args
            cmd.start = start
            cmd.end = end
            cmd.split_number = int(row['split_number'])
            cmd.sequence = scaff2sequence[scaff][start:end + 1]

            # Add to the Sdict
            Sdict[scaff + '.' + str(row['split_number'])] = None

            # Add estimated seconds
            seconds += calc_estimated_runtime(s2p[scaff]) / s2splits[scaff]
            cmds.append(cmd)

            # See if you're done
            if seconds >= SECONDS:
                cmd_groups.append(cmds)
                seconds = 0
                cmds = []

    if len(cmds) > 0:
        cmd_groups.append(cmds)

    # Make the MergeProfile command groups



    return cmd_groups, Sdict, s2splits


def calc_estimated_runtime(pairs):
    """
    Based on the number of mapped pairs, guess how long the split will take
    """
    SLOPE_CONSTANT = 0.0061401594694834305
    return (pairs * SLOPE_CONSTANT) + 0.2
